"""
open/DurusWorks/qp/pub/user.py

"""
from datetime import timedelta
from durus.persistent import PersistentObject
from durus.persistent_dict import PersistentDict
from durus.persistent_set import PersistentSet
from durus.utils import as_bytes
from qp.lib.spec import require, specify, spec, add_getters, sequence, either
from qp.lib.spec import pattern, string_classes, optional, string, datetime_with_tz
from qp.pub.common import get_publisher, get_users
from qp.lib.util import rand_str
from qp.pub.common import site_now
import sys

if sys.version < "2.6":
    from md5 import md5
else:
    from hashlib import md5

class Permissions (PersistentDict):

    data_is = {
        either(*string_classes):
        sequence(either(PersistentObject, True), PersistentSet)}

    __slots__ = ['data']

    def grant(self, permission, granter):
        require(permission, either(*string_classes))
        require(granter, either(PersistentObject, True))
        if permission not in self:
            self[permission] = PersistentSet([granter])
        else:
            self[permission].add(granter)

    def ungrant(self, permission, granter):
        require(permission, either(*string_classes))
        require(granter, either(PersistentObject, True))
        if self.is_granted(permission, granter):
            self.data[permission].remove(granter)
            if len(self.data[permission]) == 0:
                del self[permission]

    def is_granted(self, permission, granter):
        return granter in self.get(permission, [])


def compute_digest(*args):
    x = as_bytes(":".join(str(a) for a in args if a is not None))
    return md5(x).hexdigest()


class Digester (PersistentObject):

    digests_is = {either(*string_classes):either(*string_classes)}

    temp_digests_is = optional({either(*string_classes):(either(*string_classes), datetime_with_tz)})

    temp_expiration_time_is = optional(timedelta)

    __slots__ = ['digests', 'temp_digests', 'temp_expiration_time']

    def __init__(self, temp_expiration_time=None):
        self.digests = {}
        self.temp_digests = {}
        if temp_expiration_time is None:
            specify(self, temp_expiration_time=timedelta(hours=24))
        else:
            specify(self, temp_expiration_time=temp_expiration_time)

    def set_digest(self, realm, digest):
        self._p_note_change()
        self.digests[realm] = digest

    def get_digest(self, realm):
        return self.digests.get(realm)

    def get_realms(self):
        return list(self.digests.keys())

    def remove_digest(self, realm):
        self._p_note_change()
        del self.digests[realm]

    def discard_digest(self, realm):
        if realm in self.digests:
            self.remove_digest(realm)

    def get_temp_digest(self, realm):
        if hasattr(self, 'temp_digests'):
            if realm in self.temp_digests:
                return self.temp_digests[realm][0]

    def get_temp_set_time(self, realm):
        if hasattr(self, 'temp_digests'):
            if realm in self.temp_digests:
                return self.temp_digests[realm][1]

    def get_temp_expiration_time(self):
        if hasattr(self, 'temp_expiration_time'):
            return self.temp_expiration_time

    def remove_temp_digest(self, realm):
        self._p_note_change()
        del self.temp_digests[realm]

    def discard_temp_digest(self, realm):
        if hasattr(self, 'temp_digests'):
            if realm in self.temp_digests:
                self.remove_temp_digest(realm)

    def set_temp_digest(self, realm, digest):
        self._p_note_change()
        if not hasattr(self, 'temp_digests'):
            self.temp_digests = {}
        self.temp_digests[realm] = (digest, site_now())

    def is_temporary(self, realm):
        digest = self.get_digest(realm)
        temp_digest = self.get_temp_digest(realm)
        if digest and temp_digest and digest == temp_digest:
            return True
        return False

    def is_expired(self, realm):
        if self.is_temporary(realm):
            temp_set_time = self.get_temp_set_time(realm)
            temp_expiration = self.get_temp_expiration_time()
            if temp_set_time and temp_expiration and temp_set_time + temp_expiration < site_now():
                return True
        return False

    def mark_as_temporary(self, realm):
        digest = self.get_digest(realm)
        if digest:
            self.set_temp_digest(realm, digest)

    def is_permanent(self, realm):
        return not self.is_temporary()

    def mark_as_permanent(self, realm):
        self.discard_temp_digest(realm)


class TokenSet (PersistentObject):
    """
    A set of randomly generated tokens that have been delivered with forms.
    These are used to make sure that forms can't be replayed.
    """

    MAX_TOKENS = 16

    tokens_is = spec(
        [str],
        "The tokens that have been delivered with forms recently.")

    __slots__ = ['tokens']

    def __init__(self):
        self.clear()

    def clear(self):
        self.tokens = []

    def __contains__(self, token):
        return token in self.tokens

    def remove(self, token):
        self._p_note_change()
        self.tokens.remove(token)

    def discard(self, token):
        if token in self.tokens:
            self.remove(token)

    def new_token(self, bytes=8):
        token = rand_str(bytes)
        self.tokens = ([token] + self.tokens)[:self.MAX_TOKENS]
        return token


class User (PersistentObject):

    id_is = spec(
        pattern('^[-A-Za-z0-9_@.]*$'),
        "unique among users here")
    digester_is = spec(
        Digester,
        "holds password hashes")
    permissions_is = spec(
        Permissions,
        "Records permissions granted.")
    tokens_is = spec(
        TokenSet,
        "a nonce dealer")
    email_is = spec(
        (None, pattern("^.+@.+\..{2,4}$")),
        "User's email address")
    idle_session_cookie_is = optional(
        (None, string),
        "If present and not None, the cookie of an unauthenticated Session owned by this User.")

    __slots__ = ['id', 'digester', 'permissions', 'tokens', 'email', 'idle_session_cookie']

    def __init__(self, user_id, temp_expiration_time=None):
        specify(self,
                digester=self.digester_is(),
                permissions=self.permissions_is(),
                tokens=self.tokens_is(),
                email=None)
        if temp_expiration_time is not None:
            self.digester.set_temp_expiration_time(temp_expiration_time)
        if user_id is None:
            self.id = None
        else:
            specify(self, id=user_id)

    def __nonzero__(self):
        """The null user is the one with id == ''.
        """
        return self.id != ''

    __bool__ = __nonzero__

    def set_password(self, password, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        digester = self.get_digester()
        if not password:
            digester.discard_digest(realm)
        else:
            digester.set_digest(
                realm, compute_digest(self.get_id(), realm, password))

    def has_password(self, password, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        digester = self.get_digester()
        digest = digester.get_digest(realm)
        return bool(
            digest and
            digest == compute_digest(self.get_id(), realm, password))

    def is_granted(self, permission, other=True):
        """(permission:str, other:either(Persistent,True)=True) -> bool
        Does this user have the `permission` granted from the `other`?
        """
        return self.permissions.is_granted(permission, other)

    def is_admin(self):
        """() -> bool
        Has the True granted 'administrator' to this user?
        """
        return self.permissions.is_granted('administrator', True)

    def get_idle_session_cookie(self):
        return getattr(self, 'idle_session_cookie', None)

    def set_idle_session_cookie(self, value):
        specify(self, idle_session_cookie=value)

    def password_is_temporary(self, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        return self.get_digester().is_temporary(realm)

    def password_is_expired(self, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        return self.get_digester().is_expired(realm)

    def mark_password_as_temporary(self, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        return self.get_digester().mark_as_temporary(realm)

    def password_is_permanent(self, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        return self.get_digester().is_permanent(realm)

    def mark_password_as_permanent(self, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        return self.get_digester().mark_as_permanent(realm)

add_getters(User)


def get_valid_permissions(user, granter):
    if granter is True:
        return user.global_permissions
    else:
        return granter.valid_permissions

def get_users_with_permissions_for_granter(granter, permissions):
    users_with_permissions = {}
    for user in get_users().values():
        for permission in permissions:
            if user.is_granted(permission, granter):
                users_with_permissions.setdefault(user, []).append(permission)
    return users_with_permissions
